// -*- C++ -*-
/*!
 * @file  RH1FK.cpp
 * @brief RH1 FK Computaiton component
 * @date $Date$
 *
 * $Id$
 */

#include "RH1FK.h"

#include "RH1Controller/kineRH1.h"

std::vector<double> getVectorDelim(const std::string &i_line, const std::string &i_delim);

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cerr << "CRH1FK:" << x << std::endl
#define ODEBUG1(x) std::cerr << "CRH1FK:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "CRH1FK:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

// Module specification
// <rtc-template block="module_spec">
static const char* rh1fk_spec[] =
  {
    "implementation_id", "RH1FK",
    "type_name",         "RH1FK",
    "description",       "RH1 FK Computaiton component",
    "version",           "0.1",
    "vendor",            "AIST",
    "category",          "Generic",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "100.0",
    // Configuration variables
    "conf.default.CameraPos", "-0.05,0,0.05,0,0,0",
    "conf.default.RH1param", "none",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RH1FK::RH1FK(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_thetaIn("theta", m_theta),
    m_hmatOut("hmat", m_hmat),
    // </rtc-template>
    dummy(0), 
    m_numDof(DOF_RH1)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("theta", m_thetaIn);
  
  // Set OutPort buffer
  registerOutPort("hmat", m_hmatOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>
  m_kineRH1 = NULL;

}

/*!
 * @brief destructor
 */
RH1FK::~RH1FK()
{
}



RTC::ReturnCode_t RH1FK::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("CameraPos", m_CameraPos, "-0.05,0,0.05,0,0,0");
  bindParameter("RH1param", m_RH1param, "none");
  
  ODEBUG2("onInitialize(): initialization finished. ");

  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1FK::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RH1FK::onActivated(RTC::UniqueId ec_id)
{
  if(m_RH1param == "none"){
    ODEBUG1("onActivated(): RH1param should be defined ");
    return RTC::RTC_ERROR;
  }

  std::vector<double> camPos = getVectorDelim(m_CameraPos, ",");

  if(camPos.size() != 6){
    ODEBUG1("onActivated(): camera pos paramter should be dimension 6 "
	    <<" instead of "<<camPos.size());
    return RTC::RTC_ERROR;
  }

  // get camera position and rotation
  m_camera_pos = vector3(camPos[0], camPos[1], camPos[2]);
    
  m_camera_ori 
    = OpenHRP::rotFromRpy(deg2rad(camPos[3]), deg2rad(camPos[4]),
			  deg2rad(camPos[5]));

  ODEBUG2(" camera pos "<<m_camera_pos);
  ODEBUG2(" camera ori "<<m_camera_ori);

  if(m_kineRH1)
    delete m_kineRH1;

  m_kineRH1 = new CkineRH1(m_RH1param);

  if(m_kineRH1 == NULL){
    ODEBUG1("onActivated(): kineRH1 cannot be created.");
    return RTC::RTC_ERROR;
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1FK::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RH1FK::onExecute(RTC::UniqueId ec_id)
{
  // get the vector
  if(m_thetaIn.isNew()){
    m_thetaIn.read();

    if(m_theta.data.length() < m_numDof){
      ODEBUG1("onExecute(): m_theta size "<<m_theta.data.length()
	      <<" should be equal or greather than "<<m_numDof);
      return RTC::RTC_ERROR;
    }

#if DEBUG == 2
    ODEBUG2("onExecute(): input theta: ");
    for(unsigned int i=0; i<m_theta.data.length(); i++)
      std::cerr<<" "<<m_theta.data[i];
    std::cerr<<std::endl;
#endif

    std::vector<double> jointVector(m_numDof);
    for(unsigned int i=0; i<m_numDof; i++)
      jointVector[i] = m_theta.data[i];
    
    // compute FK
    if(m_kineRH1->computeFK(jointVector) == false){
      ODEBUG1("onExecute: m_kineRH1->computeFK() failed. ");
      return RTC::RTC_ERROR;
    }

    vector3 j6pos;
    matrix33 j6rot;

    if(m_kineRH1->getJointPosFK(m_numDof-1, j6pos, j6rot) == false){
      ODEBUG1("onExecute: m_kineRH1->getJointPosFK() failed. ");
      return RTC::RTC_ERROR;
    }

    // get the camera position
    vector3 absCamPos(j6pos + j6rot * m_camera_pos);
    matrix33 absCamOri(j6rot * m_camera_ori);

    m_hmat.data.length(16);
    for(unsigned int i=0; i<3; i++){
      for(unsigned int j=0; j<3; j++)
	m_hmat.data[i*4+j] = absCamOri(i, j);
      m_hmat.data[i*4+3] = absCamPos(i); 
    }
    for(unsigned int i=0; i<3; i++)
      m_hmat.data[12+i]=0.0;
    m_hmat.data[15]=1.0;

#if DEBUG == 2
    ODEBUG2("onExecute(): camera pos: "<<absCamPos);
    ODEBUG2("onExecute(): camera ori: "<<absCamOri);
    ODEBUG2("onExecute(): output value: ");
    for(unsigned int i=0; i<m_hmat.data.length(); i++)
      std::cerr<<" "<<m_hmat.data[i];
    std::cerr<<std::endl;
#endif

    m_hmatOut.write();
  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t RH1FK::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1FK::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

std::vector<double> getVectorDelim(const std::string &i_line, const std::string &i_delim)
{
  std::vector<double> vec;
  std::string::size_type begin = 0;
  std::string::size_type end = 0;
  while ( (begin = i_line.find_first_not_of(i_delim, end)) != std::string::npos ) {
    end = i_line.find_first_of(i_delim, begin);
    vec.push_back(atof(i_line.substr(begin, end-begin).c_str()));
  }
  return vec;
}


extern "C"
{
 
  void RH1FKInit(RTC::Manager* manager)
  {
    RTC::Properties profile(rh1fk_spec);
    manager->registerFactory(profile,
                             RTC::Create<RH1FK>,
                             RTC::Delete<RH1FK>);
  }
  
};


