// -*- C++ -*-
/*!
 * @file  RH1IK.cpp
 * @brief RH1 IK Computaiton component
 * @date $Date$
 *
 * $Id$
 */

#include "RH1IK.h"

#include "RH1Controller/kineRH1.h"

#define SIZE_HMAT 16
#define DOF_ENDEFF 6

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cerr << "CRH1IK:" << x << std::endl
#define ODEBUG1(x) std::cerr << "CRH1IK:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "CRH1IK:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

// Module specification
// <rtc-template block="module_spec">
static const char* rh1ik_spec[] =
  {
    "implementation_id", "RH1IK",
    "type_name",         "RH1IK",
    "description",       "RH1 IK Computaiton component",
    "version",           "0.1",
    "vendor",            "AIST",
    "category",          "Generic",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "100.0",
    // Configuration variables
    "conf.default.RH1param", "none",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RH1IK::RH1IK(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_hmatIn("hmat", m_hmat),
    m_initConfIn("initConf", m_initConf),
    m_thetaOut("theta", m_theta),

    // </rtc-template>
    dummy(0),
    m_numDof(DOF_RH1)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("hmat", m_hmatIn);
  registerInPort("initConf", m_initConfIn);
  
  // Set OutPort buffer
  registerOutPort("theta", m_thetaOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

/*!
 * @brief destructor
 */
RH1IK::~RH1IK()
{
}



RTC::ReturnCode_t RH1IK::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("RH1param", m_RH1param, "none");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1IK::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t RH1IK::onActivated(RTC::UniqueId ec_id)
{
  if(m_RH1param == "none"){
    ODEBUG1("onActivated(): RH1param should be defined ");
    return RTC::RTC_ERROR;
  }

  if(m_kineRH1)
    delete m_kineRH1;

  m_kineRH1 = new CkineRH1(m_RH1param);

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t RH1IK::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RH1IK::onExecute(RTC::UniqueId ec_id)
{
  if(m_hmatIn.isNew()){
    m_hmatIn.read();

    std::cout<<"data received: size "<<m_hmat.data.length()<<std::endl;
    for(unsigned int i=0; i<4; i++){
      for(unsigned int j=0; j<4; j++)
	std::cout<<m_hmat.data[i*4+j]<<" ";
      std::cout<<std::endl;
    }


    if(m_hmat.data.length() != SIZE_HMAT){
      ODEBUG1("onExecute(): m_hmat size "<<m_hmat.data.length()
	      <<" should be "<<SIZE_HMAT);
      return RTC::RTC_ERROR;
    }

    // get the input end-effector configuration (x, y, z, r, p, y)
    std::vector<double> goalConf(DOF_ENDEFF), jointVector;
    vector3 oriGoal, posGoal;
    matrix33 RGoal;
    for(unsigned int i=0; i<3; i++){
      for(unsigned int j=0; j<3; j++)
	RGoal(i, j) = m_hmat.data[i*4+j];
      posGoal[i] = goalConf[i] = m_hmat.data[i*4+3]; 
    }

    oriGoal = OpenHRP::rpyFromRot(RGoal);
    for(unsigned int i=0; i<3; i++)
      goalConf[3+i] = oriGoal[i];

    bool success = m_kineRH1->computeIK(goalConf, jointVector);

    if(success){

      m_theta.data.length(jointVector.size());
      for(unsigned int i=0; i<jointVector.size(); i++)
	m_theta.data[i] = jointVector[i];

      ODEBUG1("onExecute(): IK computation successful ");

    }
    else{
      ODEBUG1("onExecute(): IK computation failed for x, y, z, r, p, y:"<<posGoal<<oriGoal
	      <<"  return zero vector ");
      m_theta.data.length(0);
    }

    ODEBUG1("onExecute() output data size "<< m_theta.data.length());

    m_thetaOut.write();
  }

  if(m_initConfIn.isNew()){
    m_initConfIn.read();

    if(m_initConf.data.length() != m_numDof){
      ODEBUG1("onExecute(): m_initConf size "<<m_initConf.data.length()
	      <<" should be "<<m_numDof);
      return RTC::RTC_ERROR;
    }

    std::vector<double> jointVector(m_numDof);
    for(unsigned int i=0; i<m_numDof; i++)
      jointVector[i] = m_initConf.data[i];

    m_kineRH1->initConf(jointVector);
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1IK::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1IK::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void RH1IKInit(RTC::Manager* manager)
  {
    RTC::Properties profile(rh1ik_spec);
    manager->registerFactory(profile,
                             RTC::Create<RH1IK>,
                             RTC::Delete<RH1IK>);
  }
  
};


