// -*- C++ -*-
/*!
 * @file  RH1VController.cpp
 * @brief RH1 Virtual Controller
 * @date $Date$
 *
 * $Id$
 */

#include "RH1Controller/RH1VController.h"

#include <iostream>

#define DOF (7)
#define DOF_INPUT (7)
// #define TIMESTEP 0.001

// Module specification
// <rtc-template block="module_spec">
static const char* rh1vcontroller_spec[] =
  {
    "implementation_id", "RH1VController",
    "type_name",         "RH1VController",
    "description",       "RH1 Virtual Controller",
    "version",           "0.1",
    "vendor",            "AIST",
    "category",          "Generic",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "1.0",
    // Configuration variables
    "conf.default.RH1GainFile", "none",
    "conf.default.InputDataMode", "buffer",
    "conf.default.ComBufSizeFreq", "10",
    "conf.default.TimeStep", "0.01",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RH1VController::RH1VController(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_angleIn("angle", m_angle),
    m_thetarefIn("thetaref", m_thetaref),
    m_thetarefBufIn("thetarefBuf", m_thetarefBuf),
    m_torqueOut("torque", m_torque),
    m_thetaOut("theta", m_theta),
    m_comBufSizeOut("comBufSize", m_comBufSize),
    m_thetaFKOut("thetaFK", m_thetaFK),

    // </rtc-template>
    dummy(0),
    q(DOF),
    qold(DOF),
    q_ref(DOF),
    q_refold(DOF)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("angle", m_angleIn);
  registerInPort("thetaref", m_thetarefIn);
  registerInPort("thetarefBuf", m_thetarefBufIn);
  
  // Set OutPort buffer
  registerOutPort("torque", m_torqueOut);
  registerOutPort("theta", m_thetaOut);
  registerOutPort("comBufSize", m_comBufSizeOut);
  registerOutPort("thetaFK", m_thetaFKOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>


  m_angle.data.length(DOF);
  m_torque.data.length(DOF);
  m_thetaref.data.length(DOF_INPUT);
  m_theta.data.length(DOF_INPUT);

  Pgain = new double[DOF];
  Dgain = new double[DOF];

  for(unsigned int i = DOF_INPUT; i<DOF; i++){
    m_angle.data[i] = 0;
    m_torque.data[i] = 0;
  }

  m_comBuf.clear();

  // for debug
  std::string fname ("RH1VSeq.dat");
  std::ofstream fout(fname.c_str());
  fout.close();

  m_foutSeq.open(fname.c_str(), std::ios::out | std::ios::app);

}

/*!
 * @brief destructor
 */
RH1VController::~RH1VController()
{
}



RTC::ReturnCode_t RH1VController::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("RH1GainFile", m_RH1GainFile, "none");
  bindParameter("InputDataMode", m_InputDataMode, "buffer");
  bindParameter("ComBufSizeFreq", m_ComBufSizeFreq, "10");
  bindParameter("TimeStep", m_TimeStep, "0.01");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1VController::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1VController::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1VController::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RH1VController::onActivated(RTC::UniqueId ec_id)
{
  // std::cout<<"activating... "<<std::endl;

   if(m_RH1GainFile != "none"){
     if (access(m_RH1GainFile.c_str(), 0)){
       std::cerr << m_RH1GainFile << " not found" << std::endl;
     }
     else{
       gain.open(m_RH1GainFile.c_str());
       for (unsigned int i=0; i<DOF; i++){
	 gain >> Pgain[i];
	 gain >> Dgain[i];
	 std::cout<<i<<"th joint: Pgain "<<Pgain[i]<<", Dgain "<<Dgain[i]<<std::endl;
       }
       gain.close();
     }
   }
   else{
     std::cout<<"no gain file loaded. "<<std::endl;
   }

   for(int i=0; i < DOF; ++i){
//     qold[i] = m_angle.data[i]; // commented out by nkita 090512
     qold[i] = 0.0; // inserted by nkita 090512
   }

//   /* changed by nkita 090606
   m_angleIn.update();
   m_thetarefIn.update();
//   */
   if(m_angleIn.isNew()){
     m_angleIn.read();
     for(int i=0; i < DOF_INPUT; ++i){
       q[i] = qold[i] = m_angle.data[i];
       q_ref[i] = q_refold[i] = q[i];
       std::cout<<i<<"th joint: initial angle "<<m_angle.data[i]<<std::endl; // changed by nkita 090606
     }
   }
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1VController::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RH1VController::onExecute(RTC::UniqueId ec_id)
{
  bool angleFlag = false;
  static double dq_ref[DOF_INPUT], dq[DOF_INPUT];
   if(m_angleIn.isNew()){
    angleFlag = true;
    m_angleIn.read();
    for(int i=0; i<DOF_INPUT; i++){
      q[i] = m_angle.data[i];
      dq[i] = (q[i] - qold[i]) / m_TimeStep;

      qold[i] = q[i];
      m_theta.data[i] = q[i];
    }
    std::cout<<"angle in: ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" "<<m_angle.data[i];
    std::cout<<std::endl;
  }

  if(m_InputDataMode == "buffer"){
    // if there is new input,
    // get configuration sequence from thetarefBuf
    // add to the buffer
    if(m_thetarefBufIn.isNew()){
      m_thetarefBufIn.read();
      for(unsigned int i=0; i<m_thetarefBuf.data.length(); i++)
	m_comBuf.push_back(m_thetarefBuf.data[i].cfg);
    }

    // set the command
    if(angleFlag){
      if(m_comBuf.size() > 0){
	RTC::TimedDoubleSeq command = m_comBuf[0];
	m_comBuf.pop_front();
	for(int i=0; i<DOF_INPUT; i++)
	  q_ref[i] = command.data[i];
      }

      for(int i=0; i<DOF_INPUT; i++){
	dq_ref[i] = (q_ref[i] - q_refold[i]) / m_TimeStep;
	q_refold[i] = q_ref[i];

	// for debut
	m_foutSeq<<dq_ref[i]<<" ";

      }
      m_foutSeq<<std::endl;
    }
    
    // set the buffer size if last buffer size % m_ComBufSizeFreq == 0
    if(m_comBuf.size() % m_ComBufSizeFreq == 0){
      m_comBufSize.data = m_comBuf.size();
      m_comBufSizeOut.write();
    }
      
    // set output angle

  }
  else if (m_InputDataMode == "direct"){

    if(m_thetarefIn.isNew()){
      m_thetarefIn.read();
      for(int i=0; i<DOF_INPUT; i++){
	q_ref[i] = m_thetaref.data[i];
	dq_ref[i] = (q_ref[i] - q_refold[i]) / m_TimeStep;
	q_refold[i] = q_ref[i];
      }
      std::cout<<"thetaref in: ";
      for(int i=0; i<DOF_INPUT; i++)
	std::cout<<" "<<m_thetaref.data[i];
      std::cout<<std::endl;
    }
    /*
      m_angleIn.update();
      m_thetarefIn.update();
    */
  }
  else{
    std::cerr<<"InputDataMode "<<m_InputDataMode<<" not supported. "<<std::endl;
    return RTC::RTC_ERROR;
  }

  if(angleFlag){
    for(int i=0; i<DOF_INPUT; i++){
      m_torque.data[i] = -(q[i] - q_ref[i]) * Pgain[i] - (dq[i]- dq_ref[i]) * Dgain[i];
    }

    std::cout<<"qref: ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" "<<q_ref[i];
    std::cout<<std::endl;
    std::cout<<"dqref: ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" "<<dq_ref[i];
    std::cout<<std::endl;
    std::cout<<"dq: ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" "<<dq[i];
    std::cout<<std::endl;
    std::cout<<"torque out: ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" "<<m_torque.data[i];
    std::cout<<std::endl;
    std::cout<<"(Pgain, Dgain): ";
    for(int i=0; i<DOF_INPUT; i++)
      std::cout<<" ("<<Pgain[i]<<", "<<Dgain[i]<<")";
    std::cout<<std::endl;

    m_thetaFK = m_theta;

    m_torqueOut.write();
    // m_thetaOut.write();

    m_thetaOut.write();
    m_thetaFKOut.write();
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1VController::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t RH1VController::onError(RTC::UniqueId ec_id)
{
  std::cout<<"error occured. "<<std::endl;
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t RH1VController::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1VController::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t RH1VController::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void RH1VControllerInit(RTC::Manager* manager)
  {
    RTC::Properties profile(rh1vcontroller_spec);
    manager->registerFactory(profile,
                             RTC::Create<RH1VController>,
                             RTC::Delete<RH1VController>);
  }
  
};


